from decimal import Decimal
from markupsafe import escape, Markup
from wtforms.compat import text_type
from wtforms.widgets import html_params
from otree import settings
from otree.currency import CURRENCY_SYMBOLS
from otree.i18n import core_gettext


# the below code is adapted from wtforms


class BaseWidget:
    has_value = True

    def __call__(self, field, **render_kw):

        self.field = field
        render_kw.setdefault('id', field.id)
        if self.has_value and 'value' not in render_kw:
            render_kw['value'] = field._value()
        if 'required' not in render_kw and 'required' in getattr(field, 'flags', []):
            render_kw['required'] = True
        self.render_kw = render_kw
        return Markup(''.join(self.get_html_fragments()))

    def get_html_fragments(self):
        raise NotImplementedError

    def attrs(self):
        return html_params(name=self.field.name, **self.render_kw)


class CheckboxInput(BaseWidget):
    def __call__(self, field, **render_kw):
        if getattr(field, 'checked', field.data):
            render_kw['checked'] = True
        return super().__call__(field, **render_kw)

    def get_html_fragments(self):
        yield ('<input type="checkbox" class="form-check-input" %s>' % (self.attrs()))


class IntegerWidget(BaseWidget):
    """
    better to use number input when we can, because:
    - on mobile it pops up the number keypad
    - better validation of numbers
    """

    def get_html_fragments(self):
        min = self.render_kw.get('min')
        if min is not None and min >= 0:
            inputmode = 'numeric'
        else:
            inputmode = ''

        yield f'<input type="number" class="form-control" inputmode="{inputmode}" %s>' % self.attrs()


class FloatWidget(BaseWidget):
    """
    use input type='text' instead of 'number' because:
    - it allows comma as decimal separator
    - no annoying spinner widget
    - no problems dealing with step= argument.
    """

    def get_html_fragments(self):
        min = self.render_kw.get('min')
        if min is not None and min >= 0:
            inputmode = 'decimal'
        else:
            inputmode = ''
        yield f'<input type="text" class="form-control" inputmode="{inputmode}" %s>' % self.attrs()


class CurrencyWidget(BaseWidget):
    def __init__(self):
        if settings.USE_POINTS:
            if getattr(settings, 'POINTS_CUSTOM_NAME', None):
                CURRENCY_SYMBOL = settings.POINTS_CUSTOM_NAME
            else:
                # Translators: the label next to a "points" input field
                CURRENCY_SYMBOL = core_gettext('points')
            places = settings.POINTS_DECIMAL_PLACES
        else:
            CURRENCY_SYMBOL = CURRENCY_SYMBOLS.get(
                settings.REAL_WORLD_CURRENCY_CODE, settings.REAL_WORLD_CURRENCY_CODE
            )
            places = settings.REAL_WORLD_CURRENCY_DECIMAL_PLACES
        self.symbol = CURRENCY_SYMBOL
        self.places = places

    def get_html_fragments(self):
        yield '''<div class="input-group input-group-narrow">'''

        min = self.render_kw.get('min')
        if min is None or min < 0:
            inputmode = ''
        elif self.places == 0:
            inputmode = 'numeric'
        else:
            inputmode = 'decimal'

        if self.places == 0:
            yield f'<input type="number" class="form-control" inputmode="{inputmode}" {self.attrs()}>'
        else:
            yield f'<input type="text" class="form-control" inputmode="{inputmode}" {self.attrs()}>'
        yield f'''<span class="input-group-text">{self.symbol}</span>'''
        yield '</div>'


class TextInput(BaseWidget):
    def get_html_fragments(self):
        yield '<input type="text" class="form-control" %s>' % self.attrs()


class TextArea(BaseWidget):
    """
    Renders a multi-line text area.

    `rows` and `cols` ought to be passed as keyword args when rendering.
    """

    def get_html_fragments(self):
        yield (
            '<textarea class="form-control" %s>\r\n%s</textarea>'
            % (self.attrs(), escape(self.field._value()))
        )


class Select(BaseWidget):
    """
    Renders a select field.

    The field must provide an `iter_choices()` method which the widget will
    call on rendering; this method must yield tuples of
    `(value, label, selected)`.
    """

    has_value = False

    def get_html_fragments(self):
        yield '<select class="form-select" %s>' % self.attrs()
        yield '<option value="">--------</option>'
        for val, label, selected in self.field.iter_choices():
            yield self.render_option(val, label, selected)
        yield '</select>'

    @classmethod
    def render_option(cls, value, label, selected, **kwargs):
        if value is True:
            # Handle the special case of a 'True' value.
            value = text_type(value)
        options = dict(kwargs, value=value)
        if selected:
            options['selected'] = True
        return Markup(
            '<option %s>%s</option>' % (html_params(**options), escape(label))
        )


class SelectOption(object):
    """
    Renders the individual option from a select field.

    This is just a convenience for various custom rendering situations, and an
    option by itself does not constitute an entire field.
    """

    def __call__(self, field, **kwargs):
        return Select.render_option(
            field._value(), field.label.text, field.checked, **kwargs
        )


class RadioSelect(BaseWidget):
    has_value = False

    def get_html_fragments(self):

        yield '<div %s>' % html_params(**self.render_kw)
        for subfield in self.field:
            # 'required' attribute is missing in wtforms
            # https://github.com/wtforms/wtforms/pull/615
            # let's wait until that change gets released,
            # before fixing the case of {{ for option in field }}...
            # which unfortunately skips past this workaround
            subfield_html = (
                subfield(required=True)
                if self.render_kw.get('required')
                else subfield()
            )
            yield '<div class="form-check">%s %s</div>' % (
                subfield_html,
                subfield.label,
            )
        yield '</div>'


class RadioSelectHorizontal(BaseWidget):
    has_value = False

    def get_html_fragments(self):
        for subfield in self.field:
            subfield_html = (
                subfield(required=True)
                if self.render_kw.get('required')
                else subfield()
            )
            yield f'''
            <div class="form-check form-check-inline">
                {subfield_html}
                <label for="{subfield.id}" class="form-check-label">{subfield.label.text}</label>
            </div>
            '''


class RadioOption(BaseWidget):
    def __call__(self, field, **kwargs):
        if field.checked:
            kwargs['checked'] = True
        # see comment above about missing required attribute.
        return super().__call__(field, **kwargs)

    def get_html_fragments(self):
        yield '<input class="form-check-input" type="radio" %s>' % self.attrs()



import os
import re
import shutil
from pathlib import Path
from typing import List, Tuple

from .base import BaseCommand
from ..common import get_class_bounds

try:

    import rope.base.codeanalyze
    import rope.refactor.occurrences
    from rope.refactor import rename, move
    from rope.refactor.rename import Rename
    from rope.base.project import Project
    from rope.base.libutils import path_to_resource
    import black
except ModuleNotFoundError:
    import sys

    sys.exit(
        'Before running this command, you need to run "pip3 install -U rope black==20.8b1" '
    )
from rope.refactor.importutils import ImportTools
from collections import namedtuple
from typing import Iterable

print_function = print

MethodInfo = namedtuple('MethodInfo', ['start', 'stop', 'name', 'model'])


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument('apps', nargs='*')
        parser.add_argument(
            '--keep', action='store_true', dest='keep_old_files', default=False,
        )

    def handle(self, *args, apps, keep_old_files, **options):
        for app in Path('.').iterdir():
            if app.joinpath('models.py').exists():
                app_name = app.name
                if apps and app_name not in apps:
                    continue
                if not keep_old_files:
                    backup(app_name)
                try:
                    make_noself(app_name)
                except Exception as exc:
                    app.joinpath('__init__.py').write_text('')
                    raise
                if not keep_old_files:
                    rearrange_folder(app_name)
            # convert app.py format
            elif app.joinpath('app.py').exists():
                init = app.joinpath('__init__.py')
                init.unlink(missing_ok=True)
                app.joinpath('app.py').rename(init)
        # delete manage.py so that PyCharm doesn't try to enforce Django syntax in templates
        manage_py = Path('manage.py')
        if manage_py.exists():
            manage_py.unlink()

        print_function('Done. You should also run: otree upcase_constants')


class CannotConvert(Exception):
    pass


CURRENCY_C_IMPORT = 'Currency as c'


def make_noself(app_name):
    proj = Project(app_name, ropefolder=None)
    approot = Path(app_name)
    app_path = approot / '__init__.py'
    pages_path = approot / 'pages.py'
    models_path = approot / 'models.py'
    if not models_path.exists():
        return
    print_function('Upgrading', app_name)

    def read():
        return app_path.read_text('utf8')

    def write(txt):
        app_path.write_text(txt, encoding='utf8')

    def writelines(lines):
        write('\n'.join(lines))

    END_OF_MODELS = '"""endofmodels"""'
    lines = [
        # 2021-12-25: don't know why this is here
        'from otree.api import Page, WaitPage',
        'from otree.api import *',
        # for some reason rope considers this a duplicate import
        'from otree.api import Currency as c',
        *models_path.read_text('utf8').splitlines(),
        END_OF_MODELS,
    ]

    pages_txt = pages_path.read_text('utf8')

    m = re.search(
        r'self\.(?!player|group|subsession|participant|session|timeout_happened|round_number)(\w+)',
        pages_txt,
    )
    if m:
        msg = f"""{app_name}/pages.py contains "{m.group(0)}". This is not a recognized page attribute."""
        raise CannotConvert(msg)

    for line in pages_txt.split('\n'):
        if line.startswith('from ._builtin'):
            continue
        if line.startswith('from .models'):
            continue

        lines.append(line)

    writelines(lines)

    # normalize, get rid of empty lines
    lines = app_path.read_text('utf8').splitlines(keepends=False)
    writelines(e.replace('\t', ' ' * 4) for e in lines if e.strip())

    def resource(pth):
        return path_to_resource(proj, app_name + '/' + pth)

    app_res = resource('__init__.py')

    app_txt = read()

    # need it to be reversed so we don't shift everything down
    class_names = list(
        (m.group(1), m.group(2))
        for m in re.finditer(
            r'^class (\w+)\((BasePlayer|BaseGroup|BaseSubsession|Page|WaitPage)',
            app_txt,
            re.MULTILINE,
        )
    )

    model_methods = set()
    for class_name, base_class in reversed(class_names):
        offsets = get_method_offsets(app_txt, class_name)
        for offset, name in reversed(offsets):
            if base_class == 'WaitPage' and name == 'after_all_players_arrive':
                cls_start, cls_end = get_class_bounds(app_txt, class_name)
                is_subsession = (
                    'wait_for_all_groups = True' in app_txt[cls_start:cls_end]
                )
                rename_self_to = 'subsession' if is_subsession else 'group'
            else:
                rename_self_to = dict(
                    Player='player', Group='group', Subsession='subsession'
                ).get(class_name, 'player')
            # it might be error_message or app_after_this_page, which take extra args.
            self_offset = offset + app_txt[offset:].index('(self') + 2
            try:
                changes = Rename(proj, app_res, self_offset).get_changes(rename_self_to)
                proj.do(changes)
            except Exception:
                print_function(app_txt[self_offset : self_offset + 30])
                raise
            if class_name in ['Player', 'Group', 'Subsession']:
                model_methods.add(name)
                template_usage = f'{rename_self_to}.{name}'
                for tpl in approot.joinpath('templates', app_name).glob('*.html'):
                    if template_usage in tpl.read_text('utf8'):
                        print_function(
                            f"""
((((((((((((((((((((((((((((((
"{tpl}" contains the method call {template_usage}, but {name} has been converted to a function.
You have 2 choices:
\t(a) call {name}({rename_self_to}) in vars_for_template
\t(b) manually convert {name} back to a method
))))))))))))))))))))))))))))))"""
                        )

    app_txt = read()

    try:
        currency_offset = app_txt.index(CURRENCY_C_IMPORT)
    except ValueError:
        pass
    else:
        changes = Rename(
            proj, app_res, currency_offset + len(CURRENCY_C_IMPORT) - 1
        ).get_changes('cu')
        proj.do(changes)

    import_tools = ImportTools(proj)

    rope_module = proj.get_module('__init__')
    module_with_imports = import_tools.module_imports(rope_module)
    module_with_imports.remove_duplicates()
    module_with_imports.sort_imports()
    write(module_with_imports.get_changed_source())

    app_txt = read()
    lines = app_txt.splitlines()

    method_bounds = []
    for class_name, _ in class_names:
        if class_name in ['Player', 'Group', 'Subsession']:
            # print_function(ClassName, list(get_method_bounds(lines, ClassName)))
            method_bounds.extend(get_method_bounds(lines, class_name, start_index=0))
        else:
            for start, end, name, _ in reversed(
                list(get_method_bounds(lines, class_name, start_index=0))
            ):
                lines.insert(start, f'    @staticmethod')

    # return
    function_lines = ['# FUNCTIONS']
    non_function_lines = []

    i = 0
    for bound in method_bounds:
        non_function_lines.extend(lines[i : bound.start])
        function_lines.extend(
            dedent(line) for line in lines[bound.start : bound.stop + 1]
        )
        i = bound.stop + 1
    non_function_lines.extend(lines[i:])

    # not aapa, since we need to resolve it being defined on group vs subsession.
    for i, line in enumerate(non_function_lines):
        non_function_lines[i] = re.sub(
            r"""(live_method) = ["'](\w+)["']""", r'\1 = \2', line,
        )

    function_lines.append('# PAGES')
    function_txt = '\n'.join(function_lines)

    txt = '\n'.join(non_function_lines).replace(END_OF_MODELS, function_txt)

    txt = re.sub(r'\bplayer\.player\b', 'player', txt)
    # for AAPA

    txt = re.sub(r'\bgroup\.group\b', 'group', txt)
    txt = re.sub(r'\bsubsession\.subsession\b', 'subsession', txt)

    txt = txt.replace(
        'def before_next_page(player):',
        'def before_next_page(player, timeout_happened):',
    ).replace('player.timeout_happened', 'timeout_happened')

    # add type annotations
    # some functions have multiple args, like error_message
    txt = re.sub(r'def (\w+)\(player\b', r'def \1(player: Player', txt)
    txt = re.sub(r'def (\w+)\(group\b', r'def \1(group: Group', txt)
    txt = re.sub(r'def (\w+)\(subsession\b', r'def \1(subsession: Subsession', txt)

    txt = fix_method_calls(txt, model_methods)

    lines = txt.splitlines(keepends=False)

    # add missing 'pass' for empty classes
    lines2 = []
    for i in range(len(lines)):
        lines2.append(lines[i])
        # this will fail if the class only contains comments, but i don't see any easy solution for that.
        if lines[i].startswith('class ') and not lines[i + 1].startswith(' '):
            lines2.append(' ' * 4 + 'pass')

    write(black_format('\n'.join(lines2)))

    tests_path = approot.joinpath('tests.py')
    if tests_path.exists():

        tests_txt = tests_path.read_text('utf8')
        new_txt = (
            tests_txt.replace('from ._builtin import Bot', 'from otree.api import Bot')
            .replace('from . import pages', 'from . import *')
            .replace('from .models import Constants', '')
        )
        new_txt = re.sub(r'\bpages\.(\w)', r'\1', new_txt)
        approot.joinpath('tests_noself.py').write_text(new_txt, encoding='utf8')


def fix_method_calls(txt, model_methods):
    """this doesn't work for functions that take args. too complicated."""
    # change player.group.my_method() to my_method(player.group)
    def repl(m):
        if m.group(3) in model_methods:
            return m.group(3) + '(' + m.group(1) + m.group(2) + ')'
        return m.group()

    return re.sub(r'([\.\w]*)\b(player|group|subsession)\.(\w+)\(\)', repl, txt)


def dedent(line):
    if line.startswith(' ' * 4):
        return line[4:]
    return line


def black_format(txt):
    return black.format_str(
        txt, mode=black.Mode(line_length=100, string_normalization=False)
    )


def is_within_a_bound(bounds, lineno):
    for bound in bounds:
        if bound.start <= lineno <= bound.stop:
            return True


def get_method_bounds(lines, ModelName, start_index=1) -> Iterable[MethodInfo]:
    """1 based"""
    in_model = False

    start = None
    name = None
    model = None
    # use start=1 to match line numbers in text editor
    for lineno, line in enumerate(lines, start=start_index):
        if line.startswith(f'class {ModelName}('):
            in_model = True
            continue

        if in_model:

            if is_class_or_module_level_statement(line):
                if start:
                    yield MethodInfo(start, lineno - 1, name, model)
                    start = None
                if line.startswith('    def '):
                    start = lineno
                    m = re.search(r'def (\w+)\((\w+)', line)
                    name = m.group(1)
                    model = m.group(2)

            if is_module_level_statement(line):
                return


def is_class_or_module_level_statement(line):
    return line[:5].strip() and not line[:5].strip().startswith('#')


def is_module_level_statement(line):
    return line[:1].strip() and not line[:1].strip().startswith('#')


def get_method_offsets(txt, ClassName) -> List[Tuple[int, str]]:
    class_start, class_end = get_class_bounds(txt, ClassName)
    return [
        (m.start(), m.group(1))
        for m in re.finditer(r'^\s{4}def (\w+)\(self\b', txt, re.MULTILINE)
        if class_start < m.start() < class_end
    ]


BACKUP_FOLDER = '_REMOVE_SELF_BACKUP'


def backup(app_name):
    approot = Path(app_name)
    old_folder = Path(BACKUP_FOLDER)
    if not old_folder.exists():
        old_folder.mkdir()
    app_backup_dest = old_folder.joinpath(app_name)
    if not app_backup_dest.exists():
        shutil.copytree(approot, app_backup_dest)
    print_function(f'Your old files were saved to {BACKUP_FOLDER}/.')


def rearrange_folder(app_name):
    approot = Path(app_name)
    app_path = approot / '__init__.py'
    if not 'from otree.api' in app_path.read_text('utf8'):
        return
    print_function('Removing old files from', app_name)
    pages_path = approot / 'pages.py'
    models_path = approot / 'models.py'
    app_py_path = approot / 'app.py'
    if pages_path.exists():
        pages_path.unlink()
    if models_path.exists():
        models_path.unlink()
    if app_py_path.exists():
        app_py_path.unlink()
    _builtin = approot.joinpath('_builtin')
    if _builtin.exists():
        shutil.rmtree(_builtin)
    templates = approot.joinpath('templates', app_name)
    if templates.exists():
        copytree_py37_compat(templates, approot)
        shutil.rmtree(approot.joinpath('templates'))
    tests_noself = approot.joinpath('tests_noself.py')
    tests_path = approot.joinpath('tests.py')
    if tests_noself.exists():
        if tests_path.exists():
            tests_path.unlink()
        tests_noself.rename(tests_path)


def copytree_py37_compat(src, dst, symlinks=False, ignore=None):
    """replacement for shutil.copytree(templates, approot, dirs_exist_ok=True)"""
    for item in os.listdir(src):
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if os.path.isdir(s):
            shutil.copytree(s, d, symlinks, ignore)
        else:
            shutil.copy2(s, d)



import contextlib
import logging

from sqlalchemy import MetaData

from .base import BaseCommand
from otree.database import engine, AnyModel
from sqlalchemy.orm import close_all_sessions

logger = logging.getLogger('otree')

MSG_RESETDB_SUCCESS_FOR_HUB = 'Created new tables and columns.'
MSG_DB_ENGINE_FOR_HUB = 'Database engine'

print_function = print


class Command(BaseCommand):
    help = (
        "Resets your development database to a fresh state. "
        "All data will be deleted."
    )

    def add_arguments(self, parser):
        ahelp = (
            'Tells the resetdb command to NOT prompt the user for ' 'input of any kind.'
        )
        parser.add_argument(
            '--noinput',
            action='store_false',
            dest='interactive',
            default=True,
            help=ahelp,
        )

    def _confirm(self) -> bool:
        print_function("This will delete and recreate your database. ")
        answer = input("Proceed? (y or n): ")
        if answer:
            return answer[0].lower() == 'y'
        return False

    def handle(self, *, interactive, **options):
        if interactive and not self._confirm():
            print_function('Canceled.')
            return

        # hub depends on this string
        logger.info(f"{MSG_DB_ENGINE_FOR_HUB}: {engine.name}")

        with contextlib.closing(engine.connect()) as bind:
            trans = bind.begin()
            old_meta = MetaData()

            # these will hang if any other processes have a long-running DB connection
            old_meta.reflect(bind)
            old_meta.drop_all(bind)

            # tables get automatically created during init_orm()
            # but since we just deleted tables, we need to re-create
            AnyModel.metadata.create_all(bind)
            trans.commit()
        # oTree Hub depends on this string
        logger.info(MSG_RESETDB_SUCCESS_FOR_HUB)



import re


# Splits a string on instances of a delimiter character. Ignores quoted delimiters.
def splitc(s, delimiter, strip=False, discard_empty=False, maxsplit=-1):

    tokens, buf, expecting, escaped = [], [], None, False

    for index, char in enumerate(s):
        if expecting:
            buf.append(char)
            if char == expecting and not escaped:
                expecting = None
        else:
            if char == delimiter:
                tokens.append(''.join(buf))
                buf = []
                if len(tokens) == maxsplit:
                    buf.append(s[index+1:])
                    break
            else:
                buf.append(char)
                if char in ('"', "'"):
                    expecting = char
        escaped = not escaped if char == '\\' else False

    tokens.append(''.join(buf))

    if strip:
        tokens = [t.strip() for t in tokens]

    if discard_empty:
        tokens = [t for t in tokens if t]

    return tokens


# Splits a string using a list of regular expression patterns. Ignores quoted delimiter matches.
def splitre(s, delimiters, keepdels=False):

    tokens, buf = [], []
    end_last_match = 0

    pattern = r'''"(?:[^\\"]|\\.)*"|'(?:[^\\']|\\.)*'|%s'''
    pattern %= '|'.join(delimiters)

    for match in re.finditer(pattern, s):
        if match.group()[0] in ["'", '"']:
            buf.append(s[end_last_match:match.end()])
            end_last_match = match.end()
            continue
        buf.append(s[end_last_match:match.start()])
        tokens.append(''.join(buf))
        buf = []
        end_last_match = match.end()
        if keepdels:
            tokens.append(match.group())

    buf.append(s[end_last_match:])
    tokens.append(''.join(buf))

    return tokens



from pathlib import Path

import otree
from otree import settings
from starlette.responses import HTMLResponse

from .errors import TemplateLoadError


class FileLoader:
    def __init__(self, *dirs):
        self.dirs = dirs
        self.cache = {}

    def load(self, filename: str, template_type=None):
        if filename in self.cache:
            return self.cache[filename]

        template, path = self.load_from_disk(filename, template_type=template_type)
        self.cache[filename] = template
        return template

    def search_template(self, template_id) -> Path:
        for dir in self.dirs:
            path = Path(dir, template_id)
            if path.exists():
                return path
        msg = f"Loader cannot locate the template file '{template_id}'."
        raise TemplateLoadError(msg)

    def load_from_disk(self, template_id, template_type) -> tuple:
        from .template import Template  # todo: resolve circular import

        abspath = self.search_template(template_id)
        try:
            template_string = abspath.read_text('utf-8')
        except OSError as err:
            msg = f"FileLoader cannot load the template file '{abspath}'."
            raise TemplateLoadError(msg) from err
        template = Template(template_string, template_id, template_type=template_type)
        return template, abspath


class FileReloader(FileLoader):
    def load(self, filename: str, template_type=None):
        if filename in self.cache:
            cached_mtime, cached_path, cached_template = self.cache[filename]
            if cached_path.exists() and cached_path.stat().st_mtime == cached_mtime:
                return cached_template
        template, path = self.load_from_disk(filename, template_type=template_type)
        mtime = path.stat().st_mtime
        self.cache[filename] = (mtime, path, template)
        return template


def get_ibis_loader():
    loader_class = FileReloader if settings.DEBUG else FileLoader

    dirs = [
        Path('.'),  # for noself
        Path('_templates'),
        Path(otree.__file__).parent.joinpath('templates'),
    ] + [Path(app_name, 'templates') for app_name in settings.OTREE_APPS]
    return loader_class(*dirs)


ibis_loader = get_ibis_loader()


def get_template_name_if_exists(template_names) -> str:
    '''return the path of the first template that exists'''
    for fname in template_names:
        try:
            ibis_loader.load(fname)
        except TemplateLoadError:
            pass
        else:
            return fname
    raise TemplateLoadError(str(template_names))


def render(template_name, context, template_type=None, **extra_context):
    return HTMLResponse(
        ibis_loader.load(template_name, template_type=template_type).render(
            context, **extra_context, strict_mode=True
        )
    )
    # i used to modify the traceback to report the original error,
    # but actually i think we shouldn't.
    # The main case I had in mind was if the user calls a method like
    # player.foo(), but it's simpler if they just don't call any complex methods
    # to begin with, and just pass variables to the template.
    # that way we don't go against thre grain



import re
from importlib import import_module
from pathlib import Path

"""Don't import from remove_self, because that requires rope to be installed."""
from otree.common import get_class_bounds
from otree.constants import BaseConstants
from .base import BaseCommand

print_function = print


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument('apps', nargs='*')
        parser.add_argument(
            '--keep', action='store_true', dest='keep_old_files', default=False,
        )

    def handle(self, *args, apps, keep_old_files, **options):
        root = Path('.')
        for app in root.iterdir():
            init_path = app.joinpath('__init__.py')
            app_name = app.name
            if not init_path.exists():
                continue
            text = init_path.read_text('utf8')
            if 'class Constants(' not in text:
                print_function(f"Skipping {app_name}")
                continue
            text = text.replace('class Constants(', 'class C(')
            module = import_module(app_name)
            Constants = module.Constants
            classvars = [k for k in vars(Constants) if k not in vars(BaseConstants)]
            cls_start, cls_end = get_class_bounds(text, 'C')
            class_txt = text[cls_start:cls_end]
            for classvar in classvars:
                class_txt = re.sub(
                    r'\b' + classvar + r'\b', classvar.upper(), class_txt
                )
            final = text[:cls_start] + class_txt + text[cls_end:]
            init_path.write_text(final, encoding='utf8')

        files_to_replace = []
        # need to replace all files and templates, including in _templates folder.
        for glob in ['*.html', '*.py', '*/*.html', '*/*.py', '*/*/*.html', '*/*/*.py']:
            files_to_replace.extend(root.glob(glob))

        for p in files_to_replace:
            txt = p.read_text('utf8')
            # somehow \w also works with non-latin chars, nice
            txt = re.sub(r'\bConstants\.(\w+)\b', upcase, txt)
            p.write_text(txt, encoding='utf8')
        print_function("Done")


def upcase(match):
    upcased_name = match.group(1).upper()
    return f'C.{upcased_name}'



import os
import sys
from setuptools import setup, find_packages
import shutil
from pathlib import Path

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

import otree

version = otree.__version__

SUPPORTED_PY3_VERSIONS = [7, 8, 9, 10, 11]

# make it visible so it stands out from the rest of the spew
MSG_PY_VERSION = """
**********************************************************************************
* Error: This version of oTree is only compatible with these Python versions:
* {}
**********************************************************************************
""".format(
    ', '.join(f'3.{x}' for x in SUPPORTED_PY3_VERSIONS)
)


if sys.version_info[0] != 3 or sys.version_info[1] not in SUPPORTED_PY3_VERSIONS:
    sys.exit(MSG_PY_VERSION)


def clean_requirements(requirements_text):
    required_raw = requirements_text.splitlines()
    required = []
    for line in required_raw:
        req = line.split('#')[0].strip()
        if req:
            required.append(req)
    return required


README = Path('README.rst').read_text('utf8')
required = clean_requirements(Path('requirements.txt').read_text())


if sys.argv[-1] == 'publish':

    if Path('dist').is_dir():
        shutil.rmtree('dist')
    for cmd in [
        "python setup.py sdist",
        "twine upload dist/*",
        f'git tag -a {version} -m "version {version}"',
        "git push --tags",
    ]:
        sys.stdout.write(cmd + '\n')
        exit_code = os.system(cmd)
        if exit_code != 0:
            raise AssertionError
    if Path('build').is_dir():
        shutil.rmtree('build')

    sys.exit()


setup(
    name='otree',
    version=version,
    include_package_data=True,
    license='MIT License',
    # 2017-10-03: find_packages function works correctly, but tests
    # are still being included in the package.
    # not sure why. so instead i use
    # recursive-exclude in MANIFEST.in.
    packages=find_packages(),
    description=('Framework for multiplayer strategy games and complex surveys.'),
    long_description=README,
    url='http://otree.org/',
    author='chris@otree.org',
    author_email='chris@otree.org',
    install_requires=required,
    entry_points={'console_scripts': ['otree=otree.main:execute_from_command_line']},
    zip_safe=False,
    # we no longer need boto but people might still have [mturk] in their reqs files
    extras_require={'mturk': []},
)



import decimal
import wtforms.fields as wtfields
from otree.currency import Currency, to_dec
from otree.i18n import format_number, core_gettext
from otree import common

from . import widgets as wg


def handle_localized_number_input(val):
    if val is None:
        return val
    return val.replace(',', '.')


class FloatField(wtfields.FloatField):
    widget = wg.FloatWidget()

    def process_formdata(self, valuelist):
        if valuelist:
            try:
                self.data = float(handle_localized_number_input(valuelist[0]))
            except ValueError:
                self.data = None
                # hack: hide this from pybabel, which seems to scan for all
                # functions that end with 'gettext('.
                # wtforms already contains these translations, so they should
                # not end up in our .po file.
                _gt = self.gettext
                raise ValueError(_gt('Not a valid float value'))

    def _value(self):
        if self.data is None:
            return ''
        return format_number(self.data, places=common.FULL_DECIMAL_PLACES)


class CurrencyField(wtfields.Field):
    widget = wg.CurrencyWidget()

    def process_formdata(self, valuelist):
        if valuelist and valuelist[0]:
            try:
                data = Currency(handle_localized_number_input(valuelist[0]))
            except (decimal.InvalidOperation, ValueError):
                self.data = None
                # see the note above about gettext
                _gt = self.gettext
                raise ValueError(_gt('Not a valid decimal value'))
        else:
            data = None
        self.data = data

    def _value(self):
        if self.data is None:
            return ''
        return format_number(to_dec(self.data), places=common.FULL_DECIMAL_PLACES)


class StringField(wtfields.StringField):
    widget = wg.TextInput()


class IntegerField(wtfields.IntegerField):
    widget = wg.IntegerWidget()


def _selectfield_getitem(self, index):
    if not isinstance(index, int):
        raise IndexError
    for (i, choice) in enumerate(self):
        if index == i:
            return choice
    raise IndexError


def __iter__(self):
    """
    Add 'required' attribute to HTML:
    https://github.com/wtforms/wtforms/pull/615/files
    """
    opts = dict(
        widget=self.option_widget,
        _name=self.name,
        _form=None,
        _meta=self.meta,
        validators=self.validators,
    )
    for i, (value, label, checked) in enumerate(self.iter_choices()):
        opt = self._Option(label=label, id='%s-%d' % (self.id, i), **opts)
        opt.process(None, value)
        opt.checked = checked
        yield opt


class RadioField(wtfields.RadioField):
    widget = wg.RadioSelect()
    option_widget = wg.RadioOption()
    __getitem__ = _selectfield_getitem
    __iter__ = __iter__


class RadioFieldHorizontal(wtfields.RadioField):
    widget = wg.RadioSelectHorizontal()
    option_widget = wg.RadioOption()
    __getitem__ = _selectfield_getitem
    __iter__ = __iter__


class DropdownField(wtfields.SelectField):
    widget = wg.Select()
    option_widget = wg.SelectOption()
    __getitem__ = _selectfield_getitem
    __iter__ = __iter__


class TextAreaField(StringField):
    """
    This field represents an HTML ``<textarea>`` and can be used to take
    multi-line input.
    """

    widget = wg.TextArea()


class CheckboxField(wtfields.BooleanField):
    widget = wg.CheckboxInput()
